# The MIT License (MIT)
#
# Copyright (c) 2015 Fabian Schuh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from binascii import hexlify, unhexlify
import hashlib
import sys
import string

""" This class and the methods require python3 """
assert sys.version_info[0] == 3, "graphenelib requires python3"

""" Default Prefix """
PREFIX = "BTS"


class Base58(object) :
    """Base58 base class
    This class serves as an abstraction layer to deal with base58 encoded
    strings and their corresponding hex and binary representation throughout the
    library.
    :param data: Data to initialize object, e.g. pubkey data, address data, ...
    :type data: hex, wif, bip38 encrypted wif, base58 string
    :param str prefix: Prefix to use for Address/PubKey strings (defaults to ``BTS``)
    :return: Base58 object initialized with ``data``
    :rtype: Base58
    :raises ValueError: if data cannot be decoded
    * ``bytes(Base58)``: Returns the raw data
    * ``str(Base58)``:   Returns the readable ``Base58CheckEncoded`` data.
    * ``repr(Base58)``:  Gives the hex representation of the data.
    *  ``format(Base58,_format)`` Formats the instance according to ``_format``:
        * ``"btc"``: prefixed with ``0x80``. Yields a valid btc address
        * ``"wif"``: prefixed with ``0x00``. Yields a valid wif key
        * ``"bts"``: prefixed with ``BTS``
        * ``"muse"``: prefixed with ``MUSE``
    """
    def __init__(self, data, prefix=PREFIX) :
        self._prefix = prefix
        if all(c in string.hexdigits for c in data) :
            self._hex     = data
        elif data[0] == "5" or data[0] == "6" :
            self._hex     = base58CheckDecode(data)
        elif data[:len(self._prefix)] == self._prefix :
            self._hex     = btsBase58CheckDecode(data[len(self._prefix):])
        else :
            raise ValueError("Error loading Base58 object")

    def __format__(self, _format) :
        """ Format output according to argument _format (wif,btc,bts)
            :param str _format: Format to use
            :return: formatted data according to _format
            :rtype: str
        """
        if _format.lower() == "wif" :
            return base58CheckEncode(0x80, self._hex)
        elif _format.lower() == "encwif" :
            return base58encode(self._hex)
        elif _format.lower() == "btc":
            return base58CheckEncode(0x00, self._hex)
        elif _format.lower() == "bts" :
            return _format.upper() + str(self)
        elif _format.lower() == "bts-testnet" :
            return 'TEST' + str(self)
        elif _format.lower() == "muse" :
            return _format.upper() + str(self)
        elif _format.lower() == "steem":
            return 'STM' + str(self)
        elif _format.lower() == "test" :
            return _format.upper() + str(self)
        elif _format.lower() == "gph" :
            return _format.upper() + str(self)
        elif _format.lower() == "ppy":
            return _format.upper() + str(self)
        elif _format.lower() == "ppy-testnet":
            return 'PPYTEST' + str(self)
        else :
            sys.stderr.write("Format %s unkown. You've been warned!\n" % _format)
            return _format.upper() + str(self)

    def __repr__(self) :
        """ Returns hex value of object
            :return: Hex string of instance's data
            :rtype: hex string
        """
        return self._hex

    def __str__(self) :
        """ Return BTS-base58CheckEncoded string of data
            :return: Base58 encoded data
            :rtype: str
        """
        return btsBase58CheckEncode(self._hex)

    def __bytes__(self) :
        """ Return raw bytes
            :return: Raw bytes of instance
            :rtype: bytes
        """
        return unhexlify(self._hex)

# https://github.com/tochev/python3-cryptocoins/raw/master/cryptocoins/base58.py
BASE58_ALPHABET = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def base58decode(base58_str):
    base58_text = bytes(base58_str, "ascii")
    n = 0
    leading_zeroes_count = 0
    for b in base58_text:
        n = n * 58 + BASE58_ALPHABET.find(b)
        if n == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while n >= 256:
        div, mod = divmod(n, 256)
        res.insert(0, mod)
        n = div
    else:
        res.insert(0, n)
    return hexlify(bytearray(1) * leading_zeroes_count + res).decode('ascii')


def base58encode(hexstring):
    byteseq = bytes(unhexlify(bytes(hexstring, 'ascii')))
    n = 0
    leading_zeroes_count = 0
    for c in byteseq:
        n = n * 256 + c
        if n == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while n >= 58:
        div, mod = divmod(n, 58)
        res.insert(0, BASE58_ALPHABET[mod])
        n = div
    else:
        res.insert(0, BASE58_ALPHABET[n])
    return (BASE58_ALPHABET[0:1] * leading_zeroes_count + res).decode('ascii')


def ripemd160(s):
    ripemd160 = hashlib.new('ripemd160')
    ripemd160.update(unhexlify(s))
    return ripemd160.digest()


def doublesha256(s):
    return hashlib.sha256(hashlib.sha256(unhexlify(s)).digest()).digest()


def b58encode(v) :
    return base58encode(v)


def b58decode(v) :
    return base58decode(v)


def base58CheckEncode(version, payload):
    s = ('%.2x' % version) + payload
    checksum = doublesha256(s)[:4]
    result = s + hexlify(checksum).decode('ascii')
    return base58encode(result)


def base58CheckDecode(s):
    s   = unhexlify(base58decode(s))
    dec = hexlify(s[:-4]).decode('ascii')
    checksum = doublesha256(dec)[:4]
    assert(s[-4:] == checksum)
    return dec[2:]


def btsBase58CheckEncode(s):
    checksum = ripemd160(s)[:4]
    result = s + hexlify(checksum).decode('ascii')
    return base58encode(result)


def btsBase58CheckDecode(s):
    s   = unhexlify(base58decode(s))
    dec = hexlify(s[:-4]).decode('ascii')
    checksum = ripemd160(dec)[:4]
    assert(s[-4:] == checksum)
    return dec
